# MIT Licensed
# Copyright (c) 2024 Bert Peters <bertptrs@archlinux.org>, et al.
from __future__ import annotations
from nvchecker.api import AsyncCache, Entry, RichResult

PROJECT_INFO_URL = "https://api.launchpad.net/1.0/{launchpad}"

async def get_version(name: str, conf: Entry, *, cache: AsyncCache, **kwargs):
  launchpad = conf["launchpad"]

  project_data = await cache.get_json(PROJECT_INFO_URL.format(launchpad=launchpad))
  data = await cache.get_json(project_data['releases_collection_link'])

  return [
    RichResult(version=entry["version"], url=entry["web_link"])
    for entry in data["entries"]
  ]






# MIT licensed
# Copyright (c) 2024 Jakub Ružička <jru@debian.org>, et al.

import asyncio
import gzip
import pathlib
import urllib
from typing import Set
try:
  from compression import zstd # type: ignore
  zstd_decompressor = zstd.decompress
except ImportError:
  import zstandard
  def zstd_decompressor(data):
    dctx = zstandard.ZstdDecompressor()
    return dctx.decompress(data, max_output_size=len(data) * 30)

import lxml.etree

from nvchecker.api import session, AsyncCache, Entry, KeyManager, VersionResult


# XML namespaces used in repodata (dead links haha)
NS = {
    'common': 'http://linux.duke.edu/metadata/common',
    'repo':   'http://linux.duke.edu/metadata/repo',
    'rpm':    'http://linux.duke.edu/metadata/rpm'
}


async def get_version(
  name: str, conf: Entry, *,
  cache: AsyncCache, keymanager: KeyManager,
  **kwargs,
) -> VersionResult:
  repo = conf['repo']
  arch = conf.get('arch', 'binary')
  pkg = conf.get('pkg')
  if not pkg:
    pkg = conf.get('rpmrepo', name)

  repo_url = urllib.parse.urlparse(repo)
  repo_path = pathlib.PurePosixPath(repo_url.path)

  # get the url of repomd.xml
  repomd_path = repo_path / 'repodata' / 'repomd.xml'
  repomd_url = repo_url._replace(path=str(repomd_path)).geturl()
  # download repomd.xml (use cache)
  repomd_body = await cache.get(repomd_url, get_file) # type: ignore
  # parse repomd.xml
  repomd_xml = lxml.etree.fromstring(repomd_body)

  # get the url of *primary.xml.gz
  primary_element = repomd_xml.find('repo:data[@type="primary"]/repo:location', namespaces=NS)
  primary_path = repo_path / primary_element.get('href') # type: ignore
  primary_url = repo_url._replace(path=str(primary_path)).geturl()
  # download and decompress *primary.xml.gz (use cache)
  primary_body = await cache.get(primary_url, get_file_gz) # type: ignore
  # parse *primary.xml metadata
  metadata = lxml.etree.fromstring(primary_body)

  # use set to eliminate duplication
  versions_set: Set[str] = set()
  # iterate package metadata
  for el in metadata.findall(f'common:package[common:name="{pkg}"]', namespaces=NS):
    pkg_arch = el.findtext('common:arch', namespaces=NS)

    # filter bych arch
    if arch == 'binary':
      if pkg_arch == 'src':
        continue
    elif arch != 'any':
      if pkg_arch != arch:
        continue

    version_info = el.find('common:version', namespaces=NS)
    version = version_info.get('ver') # type: ignore
    versions_set.add(version) # type: ignore

  versions = list(versions_set)
  return versions # type: ignore


async def get_file(url: str) -> bytes:
  res = await session.get(url)
  return res.body


async def get_file_gz(url: str) -> bytes:
  res = await session.get(url)
  loop = asyncio.get_running_loop()
  if url.endswith('.gz'):
    decompressor = gzip.decompress
  elif url.endswith('.zst'):
    decompressor = zstd_decompressor
  else:
    raise Exception('unrecognized compression format', url)
  return await loop.run_in_executor(None, decompressor, res.body)



# MIT licensed
# Copyright (c) 2024 bgme <i@bgme.me>.

from lxml import html

from nvchecker.api import (
  RichResult, Entry, AsyncCache, KeyManager,
  session, GetVersionError,
)

GO_PKG_URL = 'https://pkg.go.dev/{pkg}?tab=versions'
GO_PKG_VERSION_URL = 'https://pkg.go.dev/{pkg}@{version}'


async def get_version(
    name: str, conf: Entry, *,
    cache: AsyncCache, keymanager: KeyManager,
    **kwargs,
) -> RichResult:
  key = tuple(sorted(conf.items()))
  return await cache.get(key, get_version_impl)


async def get_version_impl(info) -> RichResult:
  conf = dict(info)
  pkg_name = conf.get('go')

  url = GO_PKG_URL.format(pkg=pkg_name)
  res = await session.get(url)
  doc = html.fromstring(res.body.decode())

  elements = doc.xpath("//div[@class='Version-tag']/a/text()")
  try:
    version = elements[0] # type: ignore
    return RichResult(
      version = version, # type: ignore
      url = GO_PKG_VERSION_URL.format(pkg=pkg_name, version=version),
    )
  except IndexError:
    raise GetVersionError("parse error", pkg_name=pkg_name)



# MIT licensed
# Copyright (c) 2020 Felix Yan <felixonmars@archlinux.org>, et al.

from functools import partial

from .cmd import run_cmd

from nvchecker.api import RichResult

async def get_version(
  name, conf, *, cache, keymanager=None
):
  git = conf['git']

  use_commit = conf.get('use_commit', False)
  if use_commit:
    ref = conf.get('branch')
    if ref is None:
      ref = 'HEAD'
      gitref = None
    else:
      ref = 'refs/heads/' + ref
      gitref = ref
    cmd = f"git ls-remote {git} {ref}"
    data = await cache.get(cmd, partial(run_cmd, name))
    version = data.split(None, 1)[0]
    return RichResult(
      version = version,
      revision = version,
      gitref = gitref,
    )
  else:
    cmd = f"git ls-remote --tags --refs {git}"
    data = await cache.get(cmd, partial(run_cmd, name))
    versions = []
    for line in data.splitlines():
      revision, version = line.split("\trefs/tags/", 1)
      versions.append(RichResult(
        version = version,
        revision = revision,
        gitref = f"refs/tags/{version}",
      ))
    return versions



# MIT licensed
# Copyright (c) 2024 Rocket Aaron <i@rocka.me>, et al.

import json
import jq

from nvchecker.api import session, GetVersionError

async def get_version(name, conf, *, cache, **kwargs):
  key = tuple(sorted(conf.items()))
  return await cache.get(key, get_version_impl)

async def get_version_impl(info):
  conf = dict(info)

  try:
    program = jq.compile(conf.get('filter', '.'))
  except ValueError as e:
    raise GetVersionError('bad jq filter', exc_info=e)

  data = conf.get('post_data')
  if data is None:
    res = await session.get(conf['url'])
  else:
    res = await session.post(conf['url'], body = data, headers = {
        'Content-Type': conf.get('post_data_type', 'application/json')
      })

  try:
    obj = json.loads(res.body)
  except json.decoder.JSONDecodeError as e:
    raise GetVersionError('bad json string', exc_info=e)

  try:
    version = program.input(obj).all()
    if version == [None] and not conf.get('missing_ok', False):
      raise GetVersionError('version string not found.')
    version = [str(v) for v in version]
  except ValueError as e:
    raise GetVersionError('failed to filter json', exc_info=e)

  return version



#!/usr/bin/env python3
# MIT licensed
# Copyright (c) 2013-2024 lilydjwg <lilydjwg@gmail.com>, et al.

from __future__ import annotations

import sys
import argparse
import asyncio
from typing import Coroutine, Tuple
from pathlib import Path

import structlog

from . import core
from .util import ResultData, RawResult, KeyManager, EntryWaiter
from .ctxvars import proxy as ctx_proxy

logger = structlog.get_logger(logger_name=__name__)

def main() -> None:
  parser = argparse.ArgumentParser(description='New version checker for software')
  parser.add_argument('-k', '--keyfile',
                      metavar='FILE', type=str,
                      help='use specified keyfile (override the one in configuration file)')
  parser.add_argument('-t', '--tries', default=1, type=int, metavar='N',
                      help='try N times when network errors occur')
  parser.add_argument('--failures', action='store_true',
                      help='exit with code 3 if failures / errors happen during checking')
  parser.add_argument('-e', '--entry', type=str,
                      help='only execute on specified entry (useful for debugging)')
  core.add_common_arguments(parser)
  args = parser.parse_args()
  if core.process_common_arguments(args):
    return

  try:
    entries, options = core.load_file(
      args.file, use_keymanager=not bool(args.keyfile))

    if args.entry:
      if args.entry not in entries:
        sys.exit('Specified entry not found in config')
      entries = {args.entry: entries[args.entry]}

    if args.keyfile:
      keymanager = KeyManager(Path(args.keyfile))
    else:
      keymanager = options.keymanager
  except core.FileLoadError as e:
    sys.exit(str(e))

  if options.proxy is not None:
    ctx_proxy.set(options.proxy)

  task_sem = asyncio.Semaphore(options.max_concurrency)
  result_q: asyncio.Queue[RawResult] = asyncio.Queue()
  dispatcher = core.setup_httpclient(
    options.max_concurrency,
    options.httplib,
    options.http_timeout,
  )
  entry_waiter = EntryWaiter()
  try:
    futures = dispatcher.dispatch(
      entries, task_sem, result_q,
      keymanager, entry_waiter,
      args.tries,
      options.source_configs,
    )
  except ModuleNotFoundError as e:
    sys.exit(f'Error: {e}')

  if options.ver_files is not None:
    oldvers = core.read_verfile(options.ver_files[0])
  else:
    oldvers = {}
  result_coro = core.process_result(oldvers, result_q, entry_waiter, verbose=bool(args.entry))
  runner_coro = core.run_tasks(futures)

  if sys.version_info >= (3, 10):
    # Python 3.10 has deprecated asyncio.get_event_loop
    results, has_failures = asyncio.run(run(result_coro, runner_coro))
  else:
    # Python < 3.10 will create an eventloop when asyncio.Queue is initialized
    results, has_failures = asyncio.get_event_loop().run_until_complete(run(result_coro, runner_coro))

  if options.ver_files is not None:
    newverf = options.ver_files[1]
    if args.entry:
      # don't remove other entries when only one entry is specified on cmdline
      vers = core.read_verfile(newverf)
    else:
      vers = {}
    vers.update(results)
    core.write_verfile(newverf, vers)

  if args.failures and has_failures:
    sys.exit(3)

async def run(
  result_coro: Coroutine[None, None, Tuple[ResultData, bool]],
  runner_coro: Coroutine[None, None, None],
) -> Tuple[ResultData, bool]:
  result_fu = asyncio.create_task(result_coro)
  runner_fu = asyncio.create_task(runner_coro)
  await runner_fu
  result_fu.cancel()
  return await result_fu

if __name__ == '__main__':
  main()



# MIT licensed
# Copyright (c) 2020 lilydjwg <lilydjwg@gmail.com>, et al.
# Copyright (c) 2017 Felix Yan <felixonmars@archlinux.org>, et al.

from nvchecker.api import RichResult, GetVersionError

URL = 'https://sources.debian.org/api/src/%(pkgname)s/?suite=%(suite)s'

async def get_version(name, conf, *, cache, **kwargs):
  pkg = conf.get('debianpkg') or name
  strip_release = conf.get('strip_release', False)
  suite = conf.get('suite') or "sid"
  url = URL % {"pkgname": pkg, "suite": suite}
  data = await cache.get_json(url)

  if not data.get('versions'):
    raise GetVersionError('Debian package not found')

  r = data['versions'][0]
  if strip_release:
    version = r['version'].split("-")[0]
  else:
    version = r['version']

  return RichResult(
    version = version,
    url = f'https://sources.debian.org/src/{data["package"]}/{r["version"]}/',
  )



# MIT licensed
# Copyright (c) 2020 lilydjwg <lilydjwg@gmail.com>, et al.
# Copyright (c) 2020 Sunlei <guizaicn@gmail.com>

from xml.etree import ElementTree

from nvchecker.api import session, RichResult

XML_NAMESPACE = 'http://www.w3.org/XML/1998/namespace'
SPARKLE_NAMESPACE = 'http://www.andymatuschak.org/xml-namespaces/sparkle'

async def get_version(name, conf, *, cache, **kwargs):
  sparkle = conf['sparkle']
  release_notes_language = conf.get('release_notes_language', 'en')
  return await cache.get((sparkle, release_notes_language), get_version_impl)


async def get_version_impl(info):
  sparkle, release_notes_language = info
  res = await session.get(sparkle)
  root = ElementTree.fromstring(res.body).find('./channel/item[1]')
  item = root.find('./enclosure')

  version_string = item.get(f'{{{SPARKLE_NAMESPACE}}}shortVersionString')
  build_number = item.get(f'{{{SPARKLE_NAMESPACE}}}version')

  if (version_string and version_string.isdigit()) and (
    build_number and not build_number.isdigit()
  ):
    version_string, build_number = build_number, version_string

  version = []

  if version_string:
    version.append(version_string)
  if build_number and (build_number not in version):
    version.append(build_number)

  version_str = '-'.join(version) if version else None

  release_notes_link = None
  for release_notes in root.findall(f'./{{{SPARKLE_NAMESPACE}}}releaseNotesLink'):
    language = release_notes.get(f'{{{XML_NAMESPACE}}}lang')

    # If the release notes have no language set, store them, but keep looking for our preferred language
    if language is None:
      release_notes_link = release_notes.text.strip()

    # If the release notes match our preferred language, store them and stop looking
    if language == release_notes_language:
      release_notes_link = release_notes.text.strip()
      break

  if release_notes_link is not None:
    return RichResult(
      version = version_str,
      url = release_notes_link,
    )
  else:
    return version_str



# MIT licensed
# Copyright (c) 2025 Maxim Slipenko <maxim@slipenko.com>, et al.

from nvchecker.api import (
  GetVersionError
)
from nvchecker.httpclient.base import HTTPError

URL="https://api.snapcraft.io/v2/snaps/info/%(snap)s"

async def get_version(
  name: str, conf, *,
  cache, keymanager,
  **kwargs,
):
  try:
    snap = conf.get("snap")
    channel = conf.get("channel")

    result = await cache.get_json(
      URL % { "snap": snap },
      headers={
        "Snap-Device-Series": "16",
      },
    )
  except HTTPError:
    raise GetVersionError(f"Failed to request snap info for {snap}")
  
  for c in result['channel-map']:
    if c['channel']['name'] == channel:
      return c['version']
  
  raise GetVersionError(f"Failed to find version for {snap}")



# MIT licensed
# Copyright (c) 2013-2020 lilydjwg <lilydjwg@gmail.com>, et al.

from nvchecker.api import session, RichResult, GetVersionError

URL = 'https://archlinux.org/packages/search/json/'

async def request(pkg):
  res = await session.get(URL, params={"name": pkg})
  return res.json()

async def get_version(name, conf, *, cache, **kwargs):
  pkg = conf.get('archpkg') or name
  strip_release = conf.get('strip_release', False)
  provided = conf.get('provided')

  data = await cache.get(pkg, request)

  if not data['results']:
    raise GetVersionError('Arch package not found')

  r = [r for r in data['results'] if r['repo'] != 'testing'][0]

  if provided:
    provides = dict(x.split('=', 1) for x in r['provides'] if '=' in x)
    version = provides.get(provided, None)
    if strip_release:
      version = version.split('-', 1)[0]
  elif strip_release:
    version = r['pkgver']
  else:
    version = r['pkgver'] + '-' + r['pkgrel']

  return RichResult(
    version = version,
    url = f'https://archlinux.org/packages/{r["repo"]}/{r["arch"]}/{r["pkgname"]}/',
  )



# MIT licensed
# Copyright (c) 2013-2020 lilydjwg <lilydjwg@gmail.com>, et al.

from typing import Any, List, Union
from urllib.parse import urlencode

from nvchecker.api import VersionResult, RichResult, Entry, AsyncCache

# doc: https://developer.atlassian.com/cloud/bitbucket/rest/api-group-commits/#api-repositories-workspace-repo-slug-commits-get
BITBUCKET_URL = 'https://bitbucket.org/api/2.0/repositories/%s/commits/%s'
# doc: https://developer.atlassian.com/cloud/bitbucket/rest/api-group-refs/#api-repositories-workspace-repo-slug-refs-tags-get
BITBUCKET_MAX_TAG = 'https://bitbucket.org/api/2.0/repositories/%s/refs/tags'

async def get_version(
  name: str, conf: Entry, *,
  cache: AsyncCache,
  **kwargs: Any,
) -> VersionResult:
  repo = conf['bitbucket']
  br = conf.get('branch', '')
  use_max_tag = conf.get('use_max_tag', False)
  use_sorted_tags = conf.get('use_sorted_tags', False)

  if use_sorted_tags or use_max_tag:
    parameters = {'fields': 'values.name,values.links.html.href,next'}

    if use_sorted_tags:
      parameters['sort'] = conf.get('sort', '-target.date')
      if 'query' in conf:
        parameters['q'] = conf['query']

  if use_sorted_tags:
    url = BITBUCKET_MAX_TAG % repo
    url += '?' + urlencode(parameters)

    return await _get_tags(url, max_page=1, cache=cache)

  elif use_max_tag:
    url = BITBUCKET_MAX_TAG % repo
    url += '?' + urlencode(parameters)

    max_page = conf.get('max_page', 3)
    return await _get_tags(url, max_page=max_page, cache=cache)

  else:
    url = BITBUCKET_URL % (repo, br)
    data = await cache.get_json(url)
    return RichResult(
      version = data['values'][0]['date'].split('T', 1)[0].replace('-', ''),
      url = data['values'][0]['links']['html']['href'],
    )

async def _get_tags(
  url: str, *,
  max_page: int,
  cache: AsyncCache,
) -> VersionResult:
  ret: List[Union[str, RichResult]] = []

  for _ in range(max_page):
    data = await cache.get_json(url)
    ret.extend([
      RichResult(
        version = tag['name'],
        url = tag['links']['html']['href'],
      ) for tag in data['values']
    ])
    if 'next' in data:
      url = data['next']
    else:
      break

  return ret



# MIT licensed
# Copyright (c) 2013-2020 lilydjwg <lilydjwg@gmail.com>, et al.

import re

import structlog

from nvchecker.api import RichResult

API_URL = 'https://crates.io/api/v1/crates/%s'
# https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string
VERSION_PATTERN = r'^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$'


async def get_version(name, conf, *, cache, **kwargs):
  logger = structlog.get_logger(logger_name=__name__, name=name)
  name = conf.get('cratesio') or name
  use_pre_release = conf.get('use_pre_release', False)
  data = await cache.get_json(API_URL % name)
  results = []
  for v in data['versions']:
    if v['yanked']:
      continue
    version = v['num']
    match = re.fullmatch(VERSION_PATTERN, version)
    if match is None:
      logger.warning('ignoring invalid version', version=version)
      continue
    if not use_pre_release and match.group('prerelease'):
      continue
    results.append(
      RichResult(
        version=version,
        url=f'https://crates.io/crates/{name}/{version}',
      )
    )

  return results



# MIT licensed
# Copyright (c) 2020 Chih-Hsuan Yen <yan12125 at gmail dot com>

from typing import Dict, List, NamedTuple, Optional, Tuple
from urllib.request import parse_http_list
from urllib.parse import urljoin
import json

from nvchecker.api import session, HTTPError

class AuthInfo(NamedTuple):
  service: Optional[str]
  realm: str

def parse_www_authenticate_header(header: str) -> Tuple[str, Dict[str, str]]:
  '''
  Parse WWW-Authenticate header used in OAuth2 authentication for container
  registries. This is NOT RFC-compliant!

  Simplified from http.parse_www_authenticate_header in Werkzeug (BSD license)
  '''
  auth_type, auth_info = header.split(None, 1)
  result = {}
  for item in parse_http_list(auth_info):
    name, value = item.split("=", 1)
    if value[:1] == value[-1:] == '"':
      value = value[1:-1]
    result[name] = value
  return auth_type, result

# Inspired by https://stackoverflow.com/a/51921869
# Reference: https://github.com/containers/image/blob/v5.6.0/docker/docker_client.go

class UnsupportedAuthenticationError(NotImplementedError):
  def __init__(self):
    super().__init__('Only Bearer authentication supported for now')

async def get_registry_auth_info(registry_host: str) -> AuthInfo:
  auth_service = auth_realm = None

  try:
    await session.get(f'https://{registry_host}/v2/')
    raise UnsupportedAuthenticationError  # No authentication needed
  except HTTPError as e:
    if e.code != 401:
      raise

    auth_type, auth_info = parse_www_authenticate_header(e.response.headers['WWW-Authenticate'])
    if auth_type.lower() != 'bearer':
      raise UnsupportedAuthenticationError

    # Although 'service' is needed as per https://docs.docker.com/registry/spec/auth/token/,
    # ghcr.io (GitHub container registry) does not provide it
    auth_service = auth_info.get('service')
    auth_realm = auth_info['realm']

    return AuthInfo(auth_service, auth_realm)

async def get_container_tags(info: Tuple[str, str, AuthInfo]) -> List[str]:
  image_path, registry_host, auth_info = info
  token = await get_auth_token(auth_info, image_path)
  tags = []
  url = f'https://{registry_host}/v2/{image_path}/tags/list'

  while True:
    res = await session.get(url, headers={
      'Authorization': f'Bearer {token}',
      'Accept': 'application/json',
    })
    tags += res.json()['tags']
    link = res.headers.get('Link')
    if link is None:
      break
    else:
      url = urljoin(url, parse_next_link(link))

  return tags


async def get_auth_token(auth_info, image_path):
  auth_params = {
    'scope': f'repository:{image_path}:pull',
  }
  if auth_info.service:
    auth_params['service'] = auth_info.service
  res = await session.get(auth_info.realm, params=auth_params)
  token = res.json()['token']
  return token


def parse_next_link(value: str) -> str:
  ending = '>; rel="next"'
  if value.endswith(ending):
    return value[1:-len(ending)]
  else:
    raise ValueError(value)


async def get_container_tag_update_time(info: Tuple[str, str, str, AuthInfo]):
  '''
  Find the update time of a container tag.

  In fact, it's the creation time of the image ID referred by the tag. Tag itself does not have any update time.
  '''
  image_path, image_tag, registry_host, auth_info = info
  token = await get_auth_token(auth_info, image_path)

  # HTTP headers
  headers = {
    'Authorization': f'Bearer {token}',
    # Prefer Image Manifest Version 2, Schema 2: https://distribution.github.io/distribution/spec/manifest-v2-2/
    'Accept': ', '.join([
      'application/vnd.oci.image.manifest.v1+json',
      'application/vnd.oci.image.index.v1+json',
      'application/vnd.docker.distribution.manifest.v2+json',
      'application/vnd.docker.distribution.manifest.list.v2+json',
      'application/json',
    ]),
  }

  # Get tag manifest
  url = f'https://{registry_host}/v2/{image_path}/manifests/{image_tag}'
  res = await session.get(url, headers=headers)
  data = res.json()
  # Schema 1 returns the creation time in the response
  if data['schemaVersion'] == 1:
    return json.loads(data['history'][0]['v1Compatibility'])['created']

  # For schema 2, we have to fetch the config's blob
  # For multi-arch images, multiple manifests are bounded with the same tag. We should choose one and then request
  # the manifest's detail
  if data.get('manifests'):
    # It's quite hard to find the manifest matching with current CPU architecture and system.
    # For now we just choose the first and it should probably work for most cases
    image_digest = data['manifests'][0]['digest']
    url = f'https://{registry_host}/v2/{image_path}/manifests/{image_digest}'
    res = await session.get(url, headers=headers)
    data = res.json()

  digest = data['config']['digest']
  url = f'https://{registry_host}/v2/{image_path}/blobs/{digest}'
  res = await session.get(url, headers=headers)
  data = res.json()
  return data['created']


async def get_version(name, conf, *, cache, **kwargs):
  image_path = conf.get('container', name)
  image_tag = None
  # image tag is optional
  if ':' in image_path:
    image_path, image_tag = image_path.split(':', 1)
  registry_host = conf.get('registry', 'docker.io')
  if registry_host == 'docker.io':
    registry_host = 'registry-1.docker.io'

  auth_info = await cache.get(registry_host, get_registry_auth_info)

  # if a tag is given, return the tag's update time, otherwise return the image's tag list
  if image_tag:
    key = image_path, image_tag, registry_host, auth_info
    return await cache.get(key, get_container_tag_update_time)
  key = image_path, registry_host, auth_info
  return await cache.get(key, get_container_tags)



# MIT licensed
# Copyright (c) 2013-2020,2025 lilydjwg <lilydjwg@gmail.com>, et al.

import asyncio
from functools import partial

import structlog

from nvchecker.api import GetVersionError

async def run_cmd(name: str, cmd: str, timeout: int = 60) -> str:
  logger = structlog.get_logger(
    logger_name = __name__,
    name = name,
    cmd = cmd,
    timeout = timeout,
  )

  logger.debug('running cmd')
  p = await asyncio.create_subprocess_shell(
    cmd,
    stdout=asyncio.subprocess.PIPE,
    stderr=asyncio.subprocess.PIPE,
  )

  if hasattr(asyncio, 'timeout'):
    # Python 3.11+
    try:
      async with asyncio.timeout(timeout):
        output, error = await p.communicate()
        output_s = output.strip().decode('latin1')
        error_s = error.strip().decode(errors='replace')
    except TimeoutError:
      p.terminate()
      await p.wait()
      raise GetVersionError('command timed out', cmd=cmd)
  else:
    output, error = await p.communicate()
    output_s = output.strip().decode('latin1')
    error_s = error.strip().decode(errors='replace')

  if p.returncode != 0:
    raise GetVersionError(
      'command exited with error',
      cmd=cmd, error=error_s,
      returncode=p.returncode)
  elif not output_s:
    raise GetVersionError(
      'command exited without output',
      cmd=cmd, error=error_s,
      returncode=p.returncode)
  else:
    return output_s

async def get_version(
  name, conf, *, cache, keymanager=None
):
  cmd = conf['cmd']
  return await cache.get(cmd, partial(run_cmd, name))



# MIT licensed
# Copyright (c) 2023 Pekka Ristola <pekkarr [at] protonmail [dot] com>, et al.

from asyncio import create_subprocess_exec
from asyncio.subprocess import PIPE
import re
from typing import Tuple, List

from nvchecker.api import GetVersionError

async def get_files(info: Tuple[str, str]) -> List[str]:
  dbpath, pkg = info
  # there's no pyalpm bindings for the file databases
  cmd = ['pacman', '-Flq', '--dbpath', dbpath, pkg]

  p = await create_subprocess_exec(*cmd, stdout = PIPE, stderr = PIPE)
  stdout, stderr = await p.communicate()

  if p.returncode == 0:
    return stdout.decode().splitlines()
  else:
    raise GetVersionError(
      'pacman failed to get file list',
      pkg = pkg,
      cmd = cmd,
      stdout = stdout.decode(errors='replace'),
      stderr = stderr.decode(errors='replace'),
      returncode = p.returncode,
    )

async def get_version(name, conf, *, cache, **kwargs):
  pkg = conf['pkgname']
  repo = conf.get('repo')
  if repo is not None:
    pkg = f'{repo}/{pkg}'
  dbpath = conf.get('dbpath', '/var/lib/pacman')
  regex = re.compile(conf['filename'])
  if regex.groups > 1:
    raise GetVersionError('multi-group regex')
  strip_dir = conf.get('strip_dir', False)

  files = await cache.get((dbpath, pkg), get_files)

  for f in files:
    fn = f.rsplit('/', 1)[-1] if strip_dir else f
    match = regex.fullmatch(fn)
    if match:
      groups = match.groups()
      return groups[0] if len(groups) > 0 else fn

  raise GetVersionError('no file matches specified regex')



# MIT licensed
# Copyright (c) 2017-2020 lilydjwg <lilydjwg@gmail.com>, et al.

from nvchecker.api import RichResult

URL = 'https://release-monitoring.org/api/project/{pkg}'

async def get_version(name, conf, *, cache, **kwargs):
  pkg = conf.get('anitya_id')
  if pkg is None:
    pkg = conf.get('anitya')
  url = URL.format(pkg = pkg)
  data = await cache.get_json(url)
  return RichResult(
    version = data['version'],
    url = f'https://release-monitoring.org/project/{data["id"]}/',
  )



# vim: se sw=2:
# MIT licensed
# Copyright (c) 2013-2024 lilydjwg <lilydjwg@gmail.com>, et al.

import sys
import argparse
import shutil
import structlog
import json
import os.path

from . import core
from .util import RichResult

logger = structlog.get_logger(logger_name=__name__)

def take() -> None:
  parser = argparse.ArgumentParser(description='update version records of nvchecker')
  core.add_common_arguments(parser)
  parser.add_argument('--all', action='store_true',
                      help='take all updates')
  parser.add_argument('--ignore-nonexistent', action='store_true',
                      help='ignore nonexistent names')
  parser.add_argument('names', metavar='NAME', nargs='*',
                      help='software name to be updated. use NAME=VERSION to update '
                           'to a specific version instead of the new version.')
  args = parser.parse_args()
  if core.process_common_arguments(args):
    return

  opt = core.load_file(args.file, use_keymanager=False)[1]
  if opt.ver_files is None:
    logger.critical(
      "doesn't have 'oldver' and 'newver' set.",
      source=args.file,
    )
    sys.exit(2)
  else:
    oldverf = opt.ver_files[0]
    newverf = opt.ver_files[1]

  oldvers = core.read_verfile(oldverf)
  newvers = core.read_verfile(newverf)

  if args.all:
    oldvers.update(newvers)
  else:
    name: str
    for name in args.names:
      if "=" in name:
        name, newver = name.split("=")
        oldvers[name] = RichResult(version=newver)
      else:
        try:
          oldvers[name] = newvers[name]
        except KeyError:
          if args.ignore_nonexistent:
            logger.warning('nonexistent in newver, ignored', name=name)
            continue

          logger.critical(
            "doesn't exist in 'newver' set.", name=name,
          )
          sys.exit(2)

  try:
    if os.path.islink(oldverf):
      shutil.copy(oldverf, oldverf.with_name(oldverf.name + '~'))
    else:
      oldverf.rename(
        oldverf.with_name(oldverf.name + '~'),
      )
  except FileNotFoundError:
    pass
  core.write_verfile(oldverf, oldvers)

def cmp() -> None:
  parser = argparse.ArgumentParser(description='compare version records of nvchecker')
  core.add_common_arguments(parser)
  parser.add_argument('-j', '--json', action='store_true',
                      help='Output JSON array of dictionaries with {name, newver, oldver, [delta]} '
                           '(or array of names if --quiet)')
  parser.add_argument('-q', '--quiet', action='store_true',
                      help="Quiet mode, output only the names.")
  parser.add_argument('-a', '--all', action='store_true',
                      help="Include unchanged versions.")
  parser.add_argument('-s', '--sort',
                      choices=('parse_version', 'vercmp', 'awesomeversion', 'none'),
                      default='parse_version',
                      help='Version compare method to backwards the arrow '
                           '(default: parse_version)')
  parser.add_argument('-n', '--newer', action='store_true',
                      help='Shows only the newer ones according to --sort.')
  parser.add_argument('--exit-status', action='store_true',
                      help="exit with status 4 if there are updates")
  args = parser.parse_args()
  if core.process_common_arguments(args):
    return

  opt = core.load_file(args.file, use_keymanager=False)[1]
  if opt.ver_files is None:
    logger.critical(
      "doesn't have 'oldver' and 'newver' set.",
      source=args.file,
    )
    sys.exit(2)
  else:
    oldverf = opt.ver_files[0]
    newverf = opt.ver_files[1]

  oldvers = {k: v.version for k, v in core.read_verfile(oldverf).items()}
  newvers = {k: v.version for k, v in core.read_verfile(newverf).items()}

  differences = []

  for name, newver in sorted(newvers.items()):  # accumulate differences
    oldver = oldvers.get(name, None)

    diff = {
      'name': name,
      'oldver': oldver,
      'newver': newver
    }

    if oldver is not None and newver is not None:
      if oldver == newver:
        diff['delta'] = 'equal'

      elif args.sort == "none":
        diff['delta'] = 'new'  # assume it's a new version if we're not comparing

      else:
        from .sortversion import sort_version_keys
        version = sort_version_keys[args.sort]

        if version(oldver) > version(newver): # type: ignore
          if args.newer:
            continue  # don't store this diff
          diff['delta'] = 'old'
        else:
          diff['delta'] = 'new'

    elif oldver is None:
      diff['delta'] = 'added'

    elif newver is None:
      if args.newer:
        continue  # don't store this diff
      diff['delta'] = 'gone'

    if args.all or diff['delta'] != 'equal':
      differences.append(diff)

  if args.json:
    if args.quiet:
      print(json.dumps([diff['name'] for diff in differences], separators=(',', ':')))
    else:
      print(json.dumps(differences, sort_keys=True, separators=(',', ':')))

  elif args.quiet:
    for diff in differences:
      print(diff['name'])

  else:
    from .lib.nicelogger import Colors, support_color
    c = Colors(support_color(sys.stdout))

    diffstyles = {
      'new': {
        'symbol': '->',
        'oldc': c.red
      },
      'old': {
        'symbol': f'{c.red}<-{c.normal}',
        'oldc': c.red
      },
      'added': {
        'symbol': '++',
        'oldc': c.red
      },
      'gone': {
        'symbol': f'{c.red}--{c.normal}',
        'oldc': c.green
      },
      'equal': {
        'symbol': '==',
        'oldc': c.green
      }
    }

    for diff in differences:
      style = diffstyles[diff.get('delta', 'equal')] # type: ignore # mypy has issues with this line

      print(f'{diff["name"]} {style["oldc"]}{diff["oldver"]}{c.normal} {style["symbol"]} {c.green}{diff["newver"]}{c.normal}')

  if args.exit_status and any(
    diff.get('delta') != 'equal' for diff in differences
  ):
    sys.exit(4)



# MIT licensed
# Copyright (c) 2013-2021,2023-2024 lilydjwg <lilydjwg@gmail.com>, et al.

import structlog
from packaging.version import Version, InvalidVersion

from nvchecker.api import RichResult

async def get_version(name, conf, *, cache, **kwargs):
  logger = structlog.get_logger(logger_name=__name__, name=name)
  ret = []

  package = conf.get('pypi') or name
  use_pre_release = conf.get('use_pre_release', False)

  url = 'https://pypi.org/pypi/{}/json'.format(package)

  data = await cache.get_json(url)

  for version in data['releases'].keys():
    # Skip versions that are marked as yanked.
    if (vers := data['releases'][version]) and vers[0]['yanked']:
      continue

    try:
      parsed_version = Version(version)
    except InvalidVersion:
      if data['releases'][version]:
        # emit a warning if there is something under the invalid version
        # sympy has an empty "0.5.13-hg" version
        logger.warning('ignoring invalid version', version=version)
      continue

    if not use_pre_release and parsed_version.is_prerelease:
      continue

    ret.append(RichResult(
      version = version,
      url = f'https://pypi.org/project/{package}/{version}/',
    ))

  return ret



# MIT licensed
# Copyright (c) 2013-2020 lilydjwg <lilydjwg@gmail.com>, et al.

import asyncio
from typing import Optional, Dict

import structlog
import aiohttp

from .base import BaseSession, TemporaryError, Response, HTTPError

__all__ = ['session']

logger = structlog.get_logger(logger_name=__name__)

class AiohttpSession(BaseSession):
  session = None

  def setup(
    self,
    concurreny: int = 20,
    timeout: int = 20,
  ) -> None:
    self._concurreny = concurreny
    self._timeout = timeout

  async def request_impl(
    self, url: str, *,
    method: str,
    proxy: Optional[str] = None,
    headers: Dict[str, str] = {},
    follow_redirects: bool = True,
    params = (),
    json = None,
    body = None,
    verify_cert: bool = True,
  ) -> Response:
    if self.session is None:
      # need to create in async context
      self.session = aiohttp.ClientSession(
        connector = aiohttp.TCPConnector(limit=self._concurreny),
        timeout = aiohttp.ClientTimeout(total=self._timeout),
        trust_env = True,
      )

    kwargs = {
      'headers': headers,
      'params': params,
      'allow_redirects': follow_redirects,
    }
    if not verify_cert:
      kwargs['ssl'] = False

    if proxy is not None:
      kwargs['proxy'] = proxy
    if body is not None:
      # Make sure all backends have the same default encoding for post data.
      if 'Content-Type' not in headers:
        headers = {**headers, 'Content-Type': 'application/x-www-form-urlencoded'}
        kwargs['headers'] = headers
      kwargs['data'] = body.encode()
    elif json is not None:
      kwargs['json'] = json

    try:
      logger.debug('send request', method=method, url=url, kwargs=kwargs)
      res = await self.session.request(
        method, url, **kwargs)
    except (
      asyncio.TimeoutError, aiohttp.ClientConnectorError,
    ) as e:
      raise TemporaryError(599, repr(e), e)

    err_cls: Optional[type] = None
    if res.status >= 500:
      err_cls = TemporaryError
    elif res.status >= 400:
      err_cls = HTTPError
    if err_cls is not None:
      raise err_cls(res.status, res.reason, res)

    body = await res.content.read()
    return Response(res.headers, body)

session = AiohttpSession()



# MIT licensed
# Copyright (c) 2013-2020 lilydjwg <lilydjwg@gmail.com>, et al.

import urllib.parse

import structlog

from nvchecker.api import (
  VersionResult, RichResult, Entry,
  AsyncCache, KeyManager, TemporaryError,
)

GITLAB_URL = 'https://%s/api/v4/projects/%s/repository/commits'
GITLAB_MAX_TAG = 'https://%s/api/v4/projects/%s/repository/tags'

async def get_version(name, conf, **kwargs):
  try:
    return await get_version_real(name, conf, **kwargs)
  except TemporaryError as e:
    check_ratelimit(e, name)

async def get_version_real(
  name: str, conf: Entry, *,
  cache: AsyncCache, keymanager: KeyManager,
  **kwargs,
) -> VersionResult:
  repo = urllib.parse.quote_plus(conf['gitlab'])
  br = conf.get('branch')
  host = conf.get('host', "gitlab.com")
  use_max_tag = conf.get('use_max_tag', False)

  if use_max_tag:
    url = GITLAB_MAX_TAG % (host, repo)
  else:
    url = GITLAB_URL % (host, repo)
    if br:
      url += '?ref_name=%s' % br

  # Load token from config
  token = conf.get('token')
  # Load token from keyman
  if token is None:
    token = keymanager.get_key(host.lower(), 'gitlab_' + host.lower())

  # Set private token if token exists.
  headers = {}
  if token:
    headers["PRIVATE-TOKEN"] = token

  data = await cache.get_json(url, headers = headers)
  if use_max_tag:
    return [
      RichResult(
        version = tag['name'],
        revision = tag['commit']['id'],
        url = f'https://{host}/{conf["gitlab"]}/-/tags/{tag["name"]}',
      ) for tag in data
    ]
  else:
    return RichResult(
      version = data[0]['created_at'].split('T', 1)[0].replace('-', ''),
      revision = data[0]['id'],
      url = data[0]['web_url'],
    )

def check_ratelimit(exc, name):
  res = exc.response
  if not res:
    raise

  logger = structlog.get_logger(logger_name=__name__, name=name)
  # default -1 is used to re-raise the exception
  n = int(res.headers.get('RateLimit-Remaining', -1))
  if n == 0:
    logger.error('gitlab rate limited. Wait some time '
                 'or get an API token to increase the allowance if not yet')
  else:
    raise



# MIT licensed
# Copyright (c) 2020-2022,2024 lilydjwg <lilydjwg@gmail.com>, et al.

from typing import Dict, Optional, Tuple

import httpx

from .base import BaseSession, TemporaryError, Response, HTTPError

__all__ = ['session']

class HttpxSession(BaseSession):
  def setup(
    self,
    concurreny: int = 20,
    timeout: int = 20,
  ) -> None:
    self.clients: Dict[Tuple[Optional[str], bool], httpx.AsyncClient] = {}
    self.timeout = timeout

  async def request_impl(
    self, url: str, *,
    method: str,
    proxy: Optional[str] = None,
    headers: Dict[str, str] = {},
    follow_redirects: bool = True,
    params = (),
    json = None,
    body = None,
    verify_cert: bool = True,
  ) -> Response:
    client = self.clients.get((proxy, verify_cert))
    if not client:
      client = httpx.AsyncClient(
        timeout = httpx.Timeout(self.timeout, pool=None),
        http2 = True,
        proxy = proxy,
        verify = verify_cert,
      )
      self.clients[(proxy, verify_cert)] = client

    try:
      if body is not None:
        # Make sure all backends have the same default encoding for post data.
        if 'Content-Type' not in headers:
          headers = {**headers, 'Content-Type': 'application/x-www-form-urlencoded'}
        body = body.encode()
      r = await client.request(
        method, url, json = json, content = body,
        headers = headers,
        follow_redirects = follow_redirects,
        # httpx checks for None but not ()
        params = params or None,
      )
      err_cls: Optional[type] = None
      if r.status_code >= 500:
        err_cls = TemporaryError
      elif r.status_code >= 400:
        err_cls = HTTPError
      if err_cls is not None:
        raise err_cls(
          r.status_code,
          r.reason_phrase,
          r,
        )

    except httpx.TransportError as e:
      raise TemporaryError(599, repr(e), e)

    body = await r.aread()
    return Response(r.headers, body)

  async def aclose(self):
    for client in self.clients.values():
      await client.aclose()
    del self.clients

session = HttpxSession()



# MIT licensed
# Copyright (c) 2021 lilydjwg <lilydjwg@gmail.com>, et al.

import re

from nvchecker.api import session, GetVersionError

async def get_version(name, conf, *, cache, **kwargs):
  key = tuple(sorted(conf.items()))
  return await cache.get(key, get_version_impl)

async def get_version_impl(info):
  conf = dict(info)
  url = conf['url']
  header = conf.get('header', 'Location')
  follow_redirects = conf.get('follow_redirects', False)
  method = conf.get('method', 'HEAD')

  try:
    regex = re.compile(conf['regex'])
  except re.error as e:
    raise GetVersionError('bad regex', exc_info=e)

  res = await session.request(
    url,
    method = method,
    follow_redirects = follow_redirects,
  )

  header_value = res.headers.get(header)
  if not header_value:
    raise GetVersionError(
      'header not found or is empty',
      header = header,
      value = header_value,
    )

  try:
    version = regex.findall(header_value)
  except ValueError:
    raise GetVersionError('version string not found.')
  return version



# MIT licensed
# Copyright (c) 2020 Felix Yan <felixonmars@archlinux.org>, et al.

async def get_version(name, conf, *, cache, **kwargs):
  url = conf['mercurial'] + '/json-tags'

  data = await cache.get_json(url)

  version = [tag['tag'] for tag in data['tags']]
  return version



# MIT licensed
# Copyright (c) 2013-2020 lilydjwg <lilydjwg@gmail.com>, et al.

from nvchecker.api import RichResult

GEMS_URL = 'https://rubygems.org/api/v1/versions/%s.json'

async def get_version(name, conf, *, cache, **kwargs):
  key = conf.get('gems', name)
  data = await cache.get_json(GEMS_URL % key)
  return [
    RichResult(
      version = item['number'],
      url = f'https://rubygems.org/gems/{key}/versions/{item["number"]}',
    ) for item in data
  ]



# MIT licensed
# Copyright (c) 2013-2020 lilydjwg <lilydjwg@gmail.com>, et al.

import json as _json
from urllib.parse import urlencode
from typing import Optional, Dict, Any
import os

from tornado.httpclient import AsyncHTTPClient, HTTPRequest

try:
  import pycurl
except ImportError:
  pycurl = None # type: ignore

from .base import BaseSession, TemporaryError, Response, HTTPError

__all__ = ['session']

HTTP2_AVAILABLE = None if pycurl else False
SSL_CERT_FILE = os.environ.get('SSL_CERT_FILE')

def setup_curl(curl):
  global HTTP2_AVAILABLE
  if HTTP2_AVAILABLE is None:
    try:
      curl.setopt(pycurl.HTTP_VERSION, 4)
      HTTP2_AVAILABLE = True
    except pycurl.error:
      HTTP2_AVAILABLE = False
  elif HTTP2_AVAILABLE:
    curl.setopt(pycurl.HTTP_VERSION, 4)

  if SSL_CERT_FILE:
    curl.setopt_string(pycurl.CAINFO, SSL_CERT_FILE)
  curl.setopt_string(pycurl.ACCEPT_ENCODING, "")

class TornadoSession(BaseSession):
  def setup(
    self,
    concurreny: int = 20,
    timeout: int = 20,
  ) -> None:
    impl: Optional[str]
    if pycurl:
      impl = "tornado.curl_httpclient.CurlAsyncHTTPClient"
    else:
      impl = None
    AsyncHTTPClient.configure(
      impl, max_clients = concurreny)
    self.timeout = timeout

  async def request_impl(
    self, url: str, *,
    method: str,
    proxy: Optional[str] = None,
    headers: Dict[str, str] = {},
    follow_redirects: bool = True,
    params = (),
    json = None,
    body = None,
    verify_cert: bool = True,
  ) -> Response:
    kwargs: Dict[str, Any] = {
      'method': method,
      'headers': headers,
      'request_timeout': self.timeout,
      'follow_redirects': follow_redirects,
      'validate_cert': verify_cert,
    }

    if body:
      # By default the content type is already 'application/x-www-form-urlencoded'
      kwargs['body'] = body
    elif json:
      kwargs['body'] = _json.dumps(json)
    kwargs['prepare_curl_callback'] = setup_curl

    if proxy:
      host, port = proxy.rsplit(':', 1)
      kwargs['proxy_host'] = host
      kwargs['proxy_port'] = int(port)

    if params:
      q = urlencode(params)
      url += '?' + q

    r = HTTPRequest(url, **kwargs)
    res = await AsyncHTTPClient().fetch(
      r, raise_error=False)
    err_cls: Optional[type] = None
    if res.code >= 500:
      err_cls = TemporaryError
    elif res.code >= 400:
      err_cls = HTTPError
    if err_cls is not None:
      raise err_cls(
        res.code, res.reason, res
      )

    return Response(res.headers, res.body)

session = TornadoSession()



# MIT licensed
# Copyright (c) 2013-2020, 2024 lilydjwg <lilydjwg@gmail.com>, et al.

import time
from urllib.parse import urlencode
from typing import List, Tuple, Union, Optional
import asyncio

import structlog

from nvchecker.api import (
  VersionResult, Entry, AsyncCache, KeyManager,
  HTTPError, session, RichResult, GetVersionError,
)

ALLOW_REQUEST = None
RATE_LIMITED_ERROR = False

GITHUB_URL = 'https://api.%s/repos/%s/commits'
GITHUB_LATEST_RELEASE = 'https://api.%s/repos/%s/releases/latest'
# https://developer.github.com/v3/git/refs/#get-all-references
GITHUB_MAX_TAG = 'https://api.%s/repos/%s/git/refs/tags'
GITHUB_MAX_RELEASE = 'https://api.%s/repos/%s/releases'
GITHUB_GRAPHQL_URL = 'https://api.%s/graphql'

async def get_version(name, conf, **kwargs):
  global RATE_LIMITED_ERROR, ALLOW_REQUEST

  if RATE_LIMITED_ERROR:
    raise RuntimeError('rate limited')

  if ALLOW_REQUEST is None:
    ALLOW_REQUEST = asyncio.Event()
    ALLOW_REQUEST.set()

  for _ in range(2): # retry once
    try:
      await ALLOW_REQUEST.wait()
      return await get_version_real(name, conf, **kwargs)
    except HTTPError as e:
      if e.code in [403, 429]:
        if n := check_ratelimit(e, name):
          ALLOW_REQUEST.clear()
          await asyncio.sleep(n+1)
          ALLOW_REQUEST.set()
          continue
        RATE_LIMITED_ERROR = True
      raise

QUERY_LATEST_TAG = '''
{{
  repository(name: "{name}", owner: "{owner}") {{
    refs(refPrefix: "refs/tags/", first: 1,
         query: "{query}",
         orderBy: {{field: TAG_COMMIT_DATE, direction: DESC}}) {{
      edges {{
        node {{
          name
          target {{
            oid
          }}
        }}
      }}
    }}
  }}
}}
'''

QUERY_LATEST_RELEASE_WITH_PRERELEASES = '''
{{
  repository(name: "{name}", owner: "{owner}") {{
    releases(first: 1, orderBy: {{field: CREATED_AT, direction: DESC}}) {{
      edges {{
        node {{
          name
          url
          tag {{
            name
          }}
          tagCommit {{
            oid
          }}
        }}
      }}
    }}
  }}
}}
'''

async def get_latest_tag(key: Tuple[str, str, str, str]) -> RichResult:
  host, repo, query, token = key
  owner, reponame = repo.split('/')
  headers = {
    'Authorization': f'bearer {token}',
    'Content-Type': 'application/json',
  }
  q = QUERY_LATEST_TAG.format(
    owner = owner,
    name = reponame,
    query = query,
  )

  res = await session.post(
    GITHUB_GRAPHQL_URL % host,
    headers = headers,
    json = {'query': q},
  )
  j = res.json()

  refs = j['data']['repository']['refs']['edges']
  if not refs:
    raise GetVersionError('no tag found')

  version = refs[0]['node']['name']
  revision = refs[0]['node']['target']['oid']
  return RichResult(
    version = version,
    gitref = f"refs/tags/{version}",
    revision = revision,
    url = f'https://github.com/{repo}/releases/tag/{version}',
  )

async def get_latest_release_with_prereleases(key: Tuple[str, str, str, str]) -> RichResult:
  host, repo, token, use_release_name = key
  owner, reponame = repo.split('/')
  headers = {
    'Authorization': f'bearer {token}',
    'Content-Type': 'application/json',
  }
  q = QUERY_LATEST_RELEASE_WITH_PRERELEASES.format(
    owner = owner,
    name = reponame,
  )

  res = await session.post(
    GITHUB_GRAPHQL_URL % host,
    headers = headers,
    json = {'query': q},
  )
  j = res.json()

  refs = j['data']['repository']['releases']['edges']
  if not refs:
    raise GetVersionError('no release found')

  tag_name = refs[0]['node']['tag']['name']
  if use_release_name:
    version = refs[0]['node']['name']
  else:
    version = tag_name

  return RichResult(
    version = version,
    gitref = f"refs/tags/{tag_name}",
    revision = refs[0]['node']['tagCommit']['oid'],
    url = refs[0]['node']['url'],
  )

async def get_version_real(
  name: str, conf: Entry, *,
  cache: AsyncCache, keymanager: KeyManager,
  **kwargs,
) -> VersionResult:
  repo = conf['github']
  host = conf.get('host', "github.com")

  # Load token from config
  token = conf.get('token')
  # Load token from keyman
  if token is None:
    token = keymanager.get_key(host.lower(), 'github')

  use_latest_tag = conf.get('use_latest_tag', False)
  if use_latest_tag:
    if not token:
      raise GetVersionError('token not given but it is required')

    query = conf.get('query', '')
    return await cache.get((host, repo, query, token), get_latest_tag) # type: ignore

  use_latest_release = conf.get('use_latest_release', False)
  include_prereleases = conf.get('include_prereleases', False)
  use_release_name = conf.get('use_release_name', False)
  if use_latest_release and include_prereleases:
    if not token:
      raise GetVersionError('token not given but it is required')

    return await cache.get(
      (host, repo, token, use_release_name),
      get_latest_release_with_prereleases) # type: ignore

  br = conf.get('branch')
  path = conf.get('path')
  use_max_tag = conf.get('use_max_tag', False)
  use_max_release = conf.get('use_max_release', False)
  if use_latest_release:
    url = GITHUB_LATEST_RELEASE % (host, repo)
  elif use_max_tag:
    url = GITHUB_MAX_TAG % (host, repo)
  elif use_max_release:
    url = GITHUB_MAX_RELEASE % (host, repo)
  else:
    url = GITHUB_URL % (host, repo)
    parameters = {}
    if br:
      parameters['sha'] = br
    if path:
      parameters['path'] = path
    url += '?' + urlencode(parameters)
  headers = {
    'Accept': 'application/vnd.github.quicksilver-preview+json',
  }
  if token:
    headers['Authorization'] = f'token {token}'

  data = await cache.get_json(url, headers = headers)

  if use_max_tag:
    tags: List[Union[str, RichResult]] = [
      RichResult(
        version = ref['ref'].split('/', 2)[-1],
        gitref = ref['ref'],
        revision = ref['object']['sha'],
        url = f'https://github.com/{repo}/releases/tag/{ref["ref"].split("/", 2)[-1]}',
      ) for ref in data
    ]
    if not tags:
      raise GetVersionError('No tag found in upstream repository.')
    return tags

  if use_max_release:
    releases: List[Union[str, RichResult]] = [
      RichResult(
        version = ref['name'] if use_release_name else ref['tag_name'],
        gitref = f"refs/tags/{ref['tag_name']}",
        url = ref['html_url'],
      ) for ref in data if include_prereleases or not ref['prerelease']
    ]
    if not releases:
      raise GetVersionError('No release found in upstream repository.')
    return releases

  if use_latest_release:
    if 'tag_name' not in data:
      raise GetVersionError('No release found in upstream repository.')

    if use_release_name:
      version = data['name']
    else:
      version = data['tag_name']

    return RichResult(
      version = version,
      gitref = f"refs/tags/{data['tag_name']}",
      url = data['html_url'],
    )

  else:
    return RichResult(
      # YYYYMMDD.HHMMSS
      version = data[0]['commit']['committer']['date'].rstrip('Z').replace('-', '').replace(':', '').replace('T', '.'),
      revision = data[0]['sha'],
      url = data[0]['html_url'],
    )

def check_ratelimit(exc: HTTPError, name: str) -> Optional[int]:
  res = exc.response
  if not res:
    raise exc

  logger = structlog.get_logger(logger_name=__name__, name=name)
  if v := res.headers.get('retry-after'):
    n = int(v)
    logger.warning('retry-after', n=n)
    return n

  # default -1 is used to re-raise the exception
  n = int(res.headers.get('X-RateLimit-Remaining', -1))
  if n == 0:
    reset = int(res.headers.get('X-RateLimit-Reset'))
    logger.error(f'rate limited, resetting at {time.ctime(reset)}. '
                  'Or get an API token to increase the allowance if not yet',
                 reset = reset)
    return None

  raise exc



# MIT licensed
# Copyright (c) 2019 lilydjwg <lilydjwg@gmail.com>, et al.

from nvchecker.api import RichResult, GetVersionError

API_URL = 'https://repology.org/api/v1/project/{}'

async def get_version(name, conf, *, cache, **kwargs):
  project = conf.get('repology') or name
  repo = conf.get('repo')
  subrepo = conf.get('subrepo')
  if not repo:
    raise GetVersionError('repo field is required for repology source')

  url = API_URL.format(project)
  data = await cache.get_json(url)

  pkgs = [pkg for pkg in data if pkg['repo'] == repo]
  if not pkgs:
    raise GetVersionError('package is not found', repo=repo)

  if subrepo:
    pkgs = [pkg for pkg in pkgs if pkg.get('subrepo') == subrepo]
    if not pkgs:
        raise GetVersionError('package is not found in subrepo',
                              repo=repo, subrepo=subrepo)

  return [
    RichResult(
      version = pkg['version'],
      url = f'https://repology.org/project/{project}/packages',
    ) for pkg in pkgs
  ]



# MIT licensed
# Copyright (c) 2020 lilydjwg <lilydjwg@gmail.com>, et al.

from .httpclient import session, TemporaryError, HTTPError
from .util import (
  Entry, BaseWorker, RawResult, VersionResult, RichResult,
  AsyncCache, KeyManager, GetVersionError, EntryWaiter,
)
from .sortversion import sort_version_keys

from .ctxvars import tries, proxy, user_agent, httptoken, entry_waiter, verify_cert



# MIT licensed
# Copyright (c) 2025 Dani Rodríguez <dani@danirod.es>, et al.

from nvchecker.api import (
    Entry,
    KeyManager,
    AsyncCache,
    VersionResult,
    GetVersionError,
    HTTPError,
    session,
)
import urllib
import pathlib
from xml.etree import ElementTree


MAVEN_CENTRAL = "https://repo1.maven.org/maven2"


async def get_version(
    name: str,
    conf: Entry,
    *,
    cache: AsyncCache,
    keymanager: KeyManager,
    **kwargs,
) -> VersionResult:
    # get project coordinates
    group_id, artifact_id = conf.get("group"), conf.get("artifact")
    if not group_id:
        raise GetVersionError("Group identifier is not given")
    if not artifact_id:
        raise GetVersionError("Artifact identifier is not given")
    group_id = group_id.replace(".", "/")

    # build the URL according to the transformation rules
    # -> https://maven.apache.org/repositories/layout.html
    repo = conf.get("repo", MAVEN_CENTRAL)

    repo_url = urllib.parse.urlparse(repo)
    repo_root = pathlib.PurePosixPath(repo_url.path)
    artifact_path = repo_root / group_id / artifact_id / "maven-metadata.xml"
    artifact_url = repo_url._replace(path=str(artifact_path)).geturl()

    try:
        version = await cache.get(artifact_url, extract_metadata)
        if version is None:
            raise GetVersionError(f"Failed to get version for {group_id}:{artifact_id}")
        return version
    except HTTPError:
        raise GetVersionError(f"Failed to fetch metadata for {group_id}:{artifact_id}")


async def extract_metadata(url):
    metadata = await session.get(url)
    root = ElementTree.fromstring(metadata.body)
    release = root.find("./versioning/release")
    if release is not None:
        return release.text



# MIT licensed
# Copyright (c) 2020 Felix Yan <felixonmars@archlinux.org>, et al.

from __future__ import annotations

import re
import asyncio
from typing import Dict, Tuple
import itertools
import functools
from collections import defaultdict

from nvchecker.api import (
  session, GetVersionError, VersionResult,
  RichResult, Entry, AsyncCache, KeyManager,
)

APT_RELEASE_URL = "%s/dists/%s/Release"
APT_PACKAGES_PATH = "%s/binary-%s/Packages%s"
APT_PACKAGES_URL = "%s/dists/%s/%s"
APT_PACKAGES_SUFFIX_PREFER = (".xz", ".gz", "")

DpkgVersion = Tuple[int, str, str]

def parse_version(s: str) -> DpkgVersion:
  try:
    epoch_str, rest = s.split(':', 1)
  except ValueError:
    epoch = 0
    rest = s
  else:
    epoch = int(epoch_str)

  try:
    ver, rev = rest.split('-', 1)
  except ValueError:
    ver = rest
    rev = ''

  return epoch, ver, rev

def _compare_part(a: str, b: str) -> int:
  sa = re.split(r'(\d+)', a)
  sb = re.split(r'(\d+)', b)
  for idx, (pa, pb) in enumerate(itertools.zip_longest(sa, sb)):
    if pa is None:
      return -1
    elif pb is None:
      return 1

    if idx % 2 == 1:
      ret = int(pa) - int(pb)
      if ret != 0:
        return ret
    else:
      if pa < pb:
        return -1
      elif pa > pb:
        return 1

  return 0

def compare_version_parsed(a: DpkgVersion, b: DpkgVersion) -> int:
  ret = a[0] - b[0]
  if ret != 0:
    return ret
  ret = _compare_part(a[1], b[1])
  if ret != 0:
    return ret
  return _compare_part(a[2], b[2])

def compare_version(a: str, b: str) -> int:
  va = parse_version(a)
  vb = parse_version(b)
  return compare_version_parsed(va, vb)

def _decompress_data(url: str, data: bytes) -> str:
  if url.endswith(".xz"):
    import lzma
    data = lzma.decompress(data)
  elif url.endswith(".gz"):
    import gzip
    data = gzip.decompress(data)

  return data.decode('utf-8')

async def get_url(url: str) -> str:
  res = await session.get(url)
  data = res.body
  loop = asyncio.get_running_loop()
  return await loop.run_in_executor(
    None, _decompress_data,
    url, data)

async def parse_packages(key: Tuple[AsyncCache, str]) -> Tuple[Dict[str, str], Dict[str, str], Dict[str, str]]:
  cache, url = key
  apt_packages = await cache.get(url, get_url) # type: ignore

  pkg_map = defaultdict(list)
  srcpkg_map = defaultdict(list)
  pkg_to_src_map = defaultdict(list)

  pkg = None
  srcpkg = None
  for line in apt_packages.split('\n'):
    if line.startswith("Package: "):
      pkg = line[9:]
    elif line.startswith("Source: "):
      srcpkg = line[8:]
    elif line.startswith("Version: "):
      version = line[9:]
      if pkg is not None:
        pkg_map[pkg].append(version)
        pkg_to_src_map["%s/%s" % (pkg, version)] = srcpkg if srcpkg is not None else pkg
      if srcpkg is not None:
        srcpkg_map[srcpkg].append(version)
      pkg = srcpkg = None

  pkg_map_max = {pkg: max(vs, key=functools.cmp_to_key(compare_version))
                 for pkg, vs in pkg_map.items()}
  srcpkg_map_max = {pkg: max(vs, key=functools.cmp_to_key(compare_version))
                 for pkg, vs in srcpkg_map.items()}
  pkg_to_src_map_max = {pkg: pkg_to_src_map["%s/%s" % (pkg, vs)]
                 for pkg, vs in pkg_map_max.items()}

  return pkg_map_max, srcpkg_map_max, pkg_to_src_map_max

async def get_version(
  name: str, conf: Entry, *,
  cache: AsyncCache, keymanager: KeyManager,
  **kwargs,
) -> VersionResult:
  srcpkg = conf.get('srcpkg')
  pkg = conf.get('pkg')
  mirror = conf['mirror']
  suite = conf['suite']
  repo = conf.get('repo', 'main')
  arch = conf.get('arch', 'amd64')
  strip_release = conf.get('strip_release', False)

  if srcpkg and pkg:
    raise GetVersionError('Setting both srcpkg and pkg is ambiguous')
  elif not srcpkg and not pkg:
    pkg = name

  apt_release = await cache.get(
    APT_RELEASE_URL % (mirror, suite), get_url) # type: ignore
  for suffix in APT_PACKAGES_SUFFIX_PREFER:
    packages_path = APT_PACKAGES_PATH % (repo, arch, suffix)
    if " " + packages_path in apt_release:
      break
  else:
    raise GetVersionError('Packages file not found in APT repository')

  pkg_map, srcpkg_map, pkg_to_src_map = await cache.get(
    (cache, APT_PACKAGES_URL % (mirror, suite, packages_path)), parse_packages) # type: ignore

  if pkg and pkg in pkg_map:
    version = pkg_map[pkg]
    changelog_name = pkg_to_src_map[pkg]
  elif srcpkg and srcpkg in srcpkg_map:
    version = srcpkg_map[srcpkg]
    changelog_name = srcpkg
  else:
    raise GetVersionError('package not found in APT repository')

  # Get Changelogs field from the Release file
  changelogs_url = None
  for line in apt_release.split('\n'):
    if line.startswith('Changelogs: '):
      changelogs_url = line[12:]
      break

  # Build the changelog URL (see https://wiki.debian.org/DebianRepository/Format#Changelogs for spec)
  changelog = None
  if changelogs_url is not None and changelogs_url != 'no':
    changelog_section = changelog_name[:4] if changelog_name.startswith('lib') else changelog_name[:1]
    changelog = changelogs_url.replace('@CHANGEPATH@', f'{repo}/{changelog_section}/{changelog_name}/{changelog_name}_{version}')

  if strip_release:
    version = version.split("-")[0]

  if changelog is not None:
    return RichResult(
      version = version,
      url = changelog,
    )
  else:
    return version



# MIT licensed
# Copyright (c) 2013-2020 lilydjwg <lilydjwg@gmail.com>, et al.

from __future__ import annotations

import urllib.parse

GITEA_URL = 'https://%s/api/v1/repos/%s/commits'
GITEA_MAX_TAG = 'https://%s/api/v1/repos/%s/tags'
GITEA_LATEST_RELEASE = 'https://%s/api/v1/repos/%s/releases'

from nvchecker.api import (
  VersionResult, RichResult, Entry,
  AsyncCache, KeyManager, GetVersionError
)

async def get_version(
  name: str, conf: Entry, *,
  cache: AsyncCache, keymanager: KeyManager,
) -> VersionResult:
  repo = urllib.parse.quote(conf['gitea'])
  br = conf.get('branch')
  host = conf.get('host', 'gitea.com')
  use_max_tag = conf.get('use_max_tag', False)
  use_latest_release = conf.get('use_latest_release', False)
  if use_max_tag and use_latest_release:
      raise GetVersionError('You cannot specify both use_max_tag and use_latest_release')

  if use_max_tag:
    url = GITEA_MAX_TAG % (host, repo)
  elif use_latest_release:
    url = GITEA_LATEST_RELEASE % (host, repo)
  else:
    url = GITEA_URL % (host, repo)
    if br:
      url += '?sha=' + br

  # Load token from config
  token = conf.get('token')
  # Load token from keyman
  if token is None:
    token = keymanager.get_key(host.lower(), 'gitea_' + host.lower())

  # Set private token if token exists.
  headers = {}
  if token:
    headers["Authorization"] = f'token {token}'

  data = await cache.get_json(url, headers = headers)
  if use_max_tag:
    return [
      RichResult(
        version = tag['name'],
        revision = tag['id'],
        url = f'https://{host}/{conf["gitea"]}/releases/tag/{tag["name"]}',
      ) for tag in data
    ]
  elif use_latest_release:
    return RichResult(
      version = data[0]['name'],
      gitref = f"refs/tags/{data[0]['tag_name']}",
      url = data[0]['html_url'],
    )
  else:
    return RichResult(
      version = data[0]['commit']['committer']['date'],
      revision = data[0]['sha'],
      url = data[0]['html_url'],
    )



# MIT licensed
# Copyright (c) 2013-2020 lilydjwg <lilydjwg@gmail.com>, et al.

from nvchecker.api import RichResult

HACKAGE_URL = 'https://hackage.haskell.org/package/%s/preferred.json'

async def get_version(name, conf, *, cache, **kwargs):
  key = conf.get('hackage', name)
  data = await cache.get_json(HACKAGE_URL % key)
  version = data['normal-version'][0]
  return RichResult(
    version = version,
    url = f'https://hackage.haskell.org/package/{key}-{version}',
  )



# MIT licensed
# Copyright (c) 2013-2020 lilydjwg <lilydjwg@gmail.com>, et al.

from nvchecker_source import cmd

async def get_version(name, conf, **kwargs):
  referree = conf.get('pacman') or name
  c = "LANG=C pacman -Si %s | grep -F Version | awk '{print $3}' | head -n 1" % referree
  conf['cmd'] = c
  strip_release = conf.get('strip_release', False)

  version = await cmd.get_version(name, conf, **kwargs)

  if strip_release and '-' in version:
    version = version.rsplit('-', 1)[0]
  return version



# MIT licensed
# Copyright (c) 2024 Daniel Peukert <daniel@peukert.cc>, et al.

import asyncio
from io import BytesIO
import tarfile
from typing import List

from nvchecker.api import (
  session, VersionResult,
  Entry, AsyncCache,
  KeyManager, RichResult
)

OPAM_REPO_INDEX_URL = "%s/index.tar.gz"
OPAM_VERSION_PATH_PREFIX = "packages/%s/%s."
OPAM_VERSION_PATH_SUFFIX = "/opam"

OPAM_DEFAULT_REPO = 'https://opam.ocaml.org'
OPAM_DEFAULT_REPO_VERSION_URL = "%s/packages/%s/%s.%s"

def _decompress_and_list_files(data: bytes) -> List[str]:
  # Convert the bytes to a file object and get a list of files
  with tarfile.open(mode='r', fileobj=BytesIO(data)) as archive:
    return archive.getnames()

async def get_files(url: str) -> List[str]:
  # Download the file and get its contents
  res = await session.get(url)
  data = res.body

  # Get the file list of the archive
  loop = asyncio.get_running_loop()
  return await loop.run_in_executor(None, _decompress_and_list_files, data)

async def get_package_versions(files: List[str], pkg: str) -> List[str]:
  # Prepare the filename prefix based on the package name
  prefix = OPAM_VERSION_PATH_PREFIX % (pkg , pkg)

  # Only keep opam files that are relevant to the package we're working with
  filtered_files = []

  for filename in files:
    if filename.startswith(prefix) and filename.endswith(OPAM_VERSION_PATH_SUFFIX):
      filtered_files.append(filename[len(prefix):-1*len(OPAM_VERSION_PATH_SUFFIX)])

  return filtered_files

async def get_version(
  name: str, conf: Entry, *,
  cache: AsyncCache, keymanager: KeyManager,
  **kwargs,
):
  pkg = conf.get('pkg', name)
  repo = conf.get('repo', OPAM_DEFAULT_REPO).rstrip('/')

  # Get the list of files in the repo index (see https://opam.ocaml.org/doc/Manual.html#Repositories for repo structure)
  files = await cache.get(OPAM_REPO_INDEX_URL % repo, get_files) # type: ignore

  # Parse the version strings from the file names
  raw_versions = await get_package_versions(files, pkg)

  # Convert the version strings into RichResults
  versions = []
  for version in raw_versions:
    versions.append(RichResult(
      version = version,
      # There is no standardised URL scheme, so we only return an URL for the default registry
      url = OPAM_DEFAULT_REPO_VERSION_URL % (repo, pkg, pkg, version) if repo == OPAM_DEFAULT_REPO else None,
    ))
  return versions



# MIT licensed
# Copyright (c) 2013-2020,2024 lilydjwg <lilydjwg@gmail.com>, et al.

from datetime import datetime, timezone
import asyncio
from typing import Iterable, Dict, List, Tuple, Any, Optional

from nvchecker.api import (
  session, GetVersionError, VersionResult, RichResult,
  Entry, BaseWorker, RawResult,
)

AUR_URL = 'https://aur.archlinux.org/rpc/'

class AurResults:
  cache: Dict[str, Optional[Dict[str, Any]]]

  def __init__(self) -> None:
    self.cache = {}

  async def get_multiple(
    self,
    aurnames: Iterable[str],
  ) -> Dict[str, Optional[Dict[str, Any]]]:
    params = [('v', '5'), ('type', 'info')]
    params.extend(('arg[]', name) for name in aurnames
                  if name not in self.cache)
    res = await session.get(AUR_URL, params=params)
    data = res.json()
    new_results = {r['Name']: r for r in data['results']}

    cache = self.cache
    cache.update(new_results)
    cache.update(
      (name, None)
      for name in set(aurnames) - new_results.keys()
    )

    return {name: cache[name] for name in aurnames
            if name in cache}

class Worker(BaseWorker):
  # https://wiki.archlinux.org/index.php/Aurweb_RPC_interface#Limitations
  batch_size = 100

  async def run(self) -> None:
    tasks = self.tasks
    n_batch, left = divmod(len(tasks), self.batch_size)
    if left > 0:
      n_batch += 1

    aur_results = AurResults()

    ret = []
    for i in range(n_batch):
      s = i * self.batch_size
      batch = tasks[s : s+self.batch_size]
      fu = self._run_batch(batch, aur_results)
      ret.append(fu)

    await asyncio.gather(*ret)

  async def _run_batch(
    self,
    batch: List[Tuple[str, Entry]],
    aur_results: AurResults,
  ) -> None:
    task_by_name: Dict[str, Entry] = dict(self.tasks)

    async with self.task_sem:
      results = await _run_batch_impl(batch, aur_results)
      for name, version in results.items():
        r = RawResult(name, version, task_by_name[name])
        await self.result_q.put(r)

async def _run_batch_impl(
  batch: List[Tuple[str, Entry]],
  aur_results: AurResults,
) -> Dict[str, VersionResult]:
  aurnames = {conf.get('aur', name) for name, conf in batch}
  results = await aur_results.get_multiple(aurnames)

  ret: Dict[str, VersionResult] = {}

  for name, conf in batch:
    aurname = conf.get('aur', name)
    use_last_modified = conf.get('use_last_modified', False)
    strip_release = conf.get('strip_release', False)

    result = results.get(aurname)

    if result is None:
      ret[name] = GetVersionError('AUR upstream not found')
      continue

    version = result['Version']
    if use_last_modified:
      dt = datetime.fromtimestamp(result['LastModified'], timezone.utc)
      version += '-' + dt.strftime('%Y%m%d%H%M%S')
    if strip_release and '-' in version:
      version = version.rsplit('-', 1)[0]

    ret[name] = RichResult(
      version = version,
      url = f'https://aur.archlinux.org/packages/{aurname}',
    )

  return ret




# MIT licensed
# Copyright (c) 2020 Ypsilik <tt2laurent.maud@gmail.com>, et al.
# Copyright (c) 2013-2020 lilydjwg <lilydjwg@gmail.com>, et al.

from lxml import html, etree

from nvchecker.api import session, GetVersionError

async def get_version(name, conf, *, cache, **kwargs):
  key = tuple(sorted(conf.items()))
  return await cache.get(key, get_version_impl)

async def get_version_impl(info):
  conf = dict(info)

  encoding = conf.get('encoding')
  is_xml = conf.get('is_xml')
  if is_xml:
    parser = etree.XMLParser(encoding=encoding)
  else:
    parser = html.HTMLParser(encoding=encoding)
  data = conf.get('post_data')
  if data is None:
    res = await session.get(conf['url'])
  else:
    res = await session.post(conf['url'], body = data, headers = {
        'Content-Type': conf.get('post_data_type', 'application/x-www-form-urlencoded')
      })

  if is_xml:
    doc = etree.fromstring(res.body, base_url=conf['url'], parser=parser)
  else:
    doc = html.fromstring(res.body, base_url=conf['url'], parser=parser)

  try:
    els = doc.xpath(conf.get('xpath'))
  except ValueError:
    if not conf.get('missing_ok', False):
      raise GetVersionError('version string not found.')
  except etree.XPathEvalError as e:
    raise GetVersionError('bad xpath', exc_info=e)

  if is_xml:
    version = [
      str(el)
      if isinstance(el, str)
      else ''.join(el.itertext())
      for el in els
    ]
  else:
    version = [
      str(el)
      if isinstance(el, str)
      else str(el.text_content())
      for el in els
    ]
  return version
